#!/usr/bin/env python
#---------------------------------------------------------------
# Project         : LADS
# Module          : lads
# File            : lads.py
# Version         : $Id: lads.py,v 1.8 2005/01/02 22:28:42 fred Exp $
# Author          : Frederic Lepied
# Created On      : Sun Feb  2 18:14:21 2003
# Purpose         : daemon to monitor login/logout
#---------------------------------------------------------------

import commands        
import string
import select
import os
import sys
import syslog
import getopt
import stat
import time
import traceback

try:
    import _fam
except:
    _fam = None

# Import from lads project
from ladsconst import *
from ladslog import *
from ladsanomalies import *
import ladsconfig
import ladsaction
import ladsutmp

# Global variables

users = []
terms = {}
login_anomalies = []
logout_anomalies = []

# Options

# _debug and use_syslog are defined in ladslog
daemon = 1
config = '/etc/security/lads.conf'
timeout = 2

# Functions

def login(entry, users, terms):
    for m in login_anomalies:
        r = m(entry, users, terms)
        if r:
            for line in r:
                alert(line)
            to = ladsconfig.get_config('MAIL_TO')
            if to:
                debug('sending mail to %s' % to)
                ladsaction.send_mail(to, r)
                
    if not entry[UTMP_USER] in users:
        users.append(entry[UTMP_USER])

def logout(oentry, entry, users, terms):
    debug('logout of %s from %s' % (oentry[UTMP_USER], oentry[UTMP_HOST]))
    for m in logout_anomalies:
        r = m(oentry, entry, users, terms)
        if r:
            for line in r:
                alert(line)

            to = ladsconfig.get_config('MAIL_TO')
            if to:
                debug('sending mail to %s' % to)
                ladsaction.send_mail(to, r)

def process_entry(entry):
    check_exceptions(config)
    if entry[UTMP_TYPE] == ladsutmp.USER_PROCESS:
        login(entry, users, terms)
        terms[entry[UTMP_LINE]] = entry
    elif entry[UTMP_TYPE] == ladsutmp.DEAD_PROCESS:
        try:
            t = terms[entry[UTMP_LINE]]
            logout(t, entry, users, terms)
            del terms[entry[UTMP_LINE]]
        except KeyError:
            alert('Unknown user log out from %s' % entry[UTMP_LINE])

def log_stacktrace():
    map(alert, traceback.format_exception(*sys.exc_info()))
    
# Anomaly detection functions

login_anomalies.append(first_login_anomaly)
login_anomalies.append(multiple_login_anomaly)
login_anomalies.append(foreign_login_anomaly)
login_anomalies.append(country_login_anomaly)

# Main part

# Process options

try:
    (opt, args)=getopt.getopt(sys.argv[1:],
                              'dfec:p',
                              ['debug',
                               'error',
                               'foreground',
                               'config=',
                               'polling'
                               ])
except getopt.error:
    print 'bad option'
    sys.exit(1)

for o in opt:
    if o[0] == '-d' or o[0] == '--debug':
        set_debug(1)
    elif o[0] == '-e' or o[0] == '--error':
        set_use_syslog(0)
    elif o[0] == '-f' or o[0] == '--foreground':
        daemon = 0
    elif o[0] == '-c' or o[0] == '--config':
        config = o[1]
    elif o[0] == '-p' or o[0] == '--polling':
        _fam = None
        
ladsconfig.load_exceptions(config)

# Daemonize

if daemon:
    if os.fork() != 0:
        sys.exit(0)
    os.chdir('/')
    for i in range(0, 3):
        os.close(i)
    os.open('/dev/null', os.O_RDONLY)
    os.open('/dev/null', os.O_WRONLY)
    os.open('/dev/null', os.O_WRONLY)
    os.setsid()
    use_syslog = 1
    # Store the pid if possible
    pidfile = '/var/run/lads.pid'
    try:
        fd = open(pidfile, 'w')
        fd.write('%d' % os.getpid())
        fd.close()
        debug('pid stored in %s' % pidfile)
    except:
        log('Unable to store pid in %s: %s' % (pidfile, str(sys.exc_value)))
    
# Fill the users that have already been connected once with info from lastlog

r = commands.getstatusoutput("LC_ALL=C lastlog | fgrep -v '**Never logged in**' | cut -f1 -d ' '")

if r[0] == 0:
    for u in string.split(r[1], '\n')[1:]:
        users.append(u)

# Fill the users already connected

wtmp = ladsutmp.utmpEntries(ladsutmp.WTMP_FILE)

entry = wtmp.get_entry()

while entry:
    if entry[UTMP_TYPE] == ladsutmp.USER_PROCESS:
        terms[entry[UTMP_LINE]] = entry
    elif entry[UTMP_TYPE] == ladsutmp.DEAD_PROCESS:
        try:
            del terms[entry[UTMP_LINE]]
        except KeyError:
            pass
    entry = wtmp.get_entry()

# Process connected users

# debug('Processing connected users')
# for t in terms.keys():
#     entry = terms[t]
    
# Main loop

try:
    fc = _fam.open()
    debug('Monitoring wtmp using FAM')
except:
    fc = None
    log('FAM not available, monitoring wtmp via polling')

if fc:
    wtmp_req = fc.monitorFile(ladsutmp.WTMP_FILE, None)
else:
    size = os.fstat(wtmp.fileno())[stat.ST_SIZE]

entry = None

try:
    while 1:
        if fc:
            try:
                ri, ro, re = select.select([fc], [], [])
            except select.error, er:
                errnumber, strerr = er
                if errnumber == errno.EINTR:
                    continue
                else:
                    log_stacktrace()
                    sys.exit(1)
        
            # FAM event
            while fc.pending():
                fe = fc.nextEvent()
                if fe.code == _fam.Deleted:
                    debug('Closing wtmp')
                    wtmp.close()
                elif fe.code == _fam.Created:
                    debug('Opening wtmp')
                    wtmp = ladsutmp.utmpEntries(ladsutmp.WTMP_FILE)
                    files = [wtmp, fc]
                elif fe.code == _fam.Changed:
                    entry = wtmp.get_entry()
                    if entry:
                        process_entry(entry)
                    
        else:    
            # check that the file hasn't been rotated
            fdstat = os.fstat(wtmp.fileno())
            try:
                namestat = os.stat(ladsutmp.WTMP_FILE)
            except OSError:
                namestat = None
            
            if not namestat or abs(fdstat[stat.ST_MTIME] - namestat[stat.ST_MTIME]) >= 10:
                log('Reopening wtmp')
                wtmp.close()
                wtmp = ladsutmp.utmpEntries(ladsutmp.WTMP_FILE)
            
            if fdstat[stat.ST_SIZE] > size:
                entry = wtmp.get_entry()
            
            size = fdstat[stat.ST_SIZE]
            
            if not entry:
                time.sleep(timeout)
                continue
            
            process_entry(entry)
            
            entry = None
except:
    log_stacktrace()

# lads.py ends here
